"""  aidgaf-server - A simple server to handle requests from the aidgaf client. 
    This server will communicate with OpenAI to generate responses to the client."""
from http.server import BaseHTTPRequestHandler, HTTPServer
import json
import threading
import requests

import merge
import idgaf
import settings
import security
from const import UTF8, REPORTING_TIMEOUT

default_request_body = b'{"message":{"command":"aidgaf","data":{"username":"AdamOutler"}}}'
request_headers = {}


class IDGAFServer(BaseHTTPRequestHandler):
    """ This class handles the requests from the client. """

    def do_GET(self):
        """ This function handles GET requests. """
        self.send_response(418)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(bytes("HTTP 408\nI'm a teapot", UTF8))
        self.close_connection

    def do_PATCH(self):
        """ This function handles PATCH requests. """
        body = self.get_body().decode(UTF8)
        check_result = perform_sanity_checks(body)
        if check_result:
            self.send_response(403)
            self.end_headers()
            self.wfile.write(bytes(check_result, UTF8))
            self.close_connection
            return
        command = json.loads(body)
        self.do_request_handling(command)

    def do_request_handling(self, command):
        """ This function handles the request. 
            Parameters:
                command: The command object received from the client. """
        print(command)
        if command['message']['command'] == 'aidgaf':
            self.idgaf_command_handler(command)

    def idgaf_command_handler(self, command):
        """ This function handles the idgaf command.  It will generate a 
            response and send it to the client.
            Parameters:
                command: The command object received from the client. """
        if not settings.ASYNC_URL:
            [code, body] = self.getGPTResponse(command)
            self.handle_response(code, body)
            return

        thread = threading.Thread(
            target=self.send_async_request, args=(command,))
        thread.start()
        self.handle_response(200, json.dumps(command))

    def send_async_request(self, command):
        """ This function sends the request to the async URL.  It will generate a
            response and send it to the client. This function is called in a separate
            thread.  It will not block the main thread.  

            Parameters:
                command: The command object received from the client. """
        print("using async for response")
        [code, body] = self.getGPTResponse(command)
        merged = merge.merge_dict_no_overwrite(json.loads(body), command)
        merged['hash'] = security.get_message_hash(json.dumps(merged))
        merged['service'] = 'JavaShark'
        data = json.dumps(merged)
        print(str(code)+": "+data)

        if (code != 200):
            return
        if (settings.ASYNC_METHOD == "POST"):
            requests.post(settings.ASYNC_URL, data=data, headers=request_headers,
                          timeout=REPORTING_TIMEOUT)
        elif (settings.ASYNC_METHOD == "PUT"):
            requests.put(settings.ASYNC_URL, data=data, headers=request_headers,
                         timeout=REPORTING_TIMEOUT)
        elif (settings.ASYNC_METHOD == "PATCH"):
            resp = requests.patch(settings.ASYNC_URL, data=data, headers=request_headers,
                                  timeout=REPORTING_TIMEOUT)
            print(str(resp)+resp.text)

    # gpt_response = requests.post(URL, json=data, headers=request_headers)

    def getGPTResponse(self, command) -> [int, str]:
        """ This function generates a response using OpenAI.  It will return 
            the response code and the response body.
            Parameters: 
                command: The command object received from the client.
            Returns:
                The response code and the response body. """
        [responseCode, json_response] = idgaf.parse_idgaf_request(command)
        json_response['hash'] = security.get_message_hash(json.dumps(command))
        response_body = json.dumps(json_response)
        return [responseCode, response_body]

    def get_body(self) -> bytes:
        """ This function returns the body of the request.
            Returns:
                The body of the request.
            """
        header_length = self.headers.get('Content-Length')
        request_body = default_request_body
        if header_length != None and self.headers.get('Content-Length') != None:
            request_body = self.rfile.read(
                int(self.headers.get('Content-Length')))
        str_request = str(request_body, UTF8)
        print(request_body)
        return request_body

    def handle_response(self, code, body):
        """ This function handles the response to the client.
            Parameters:
                code: The HTTP response code.
                body: The body of the response. """
        self.send_response(code)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        print("sending:"+body)
        self.wfile.write(bytes(body, UTF8))
        self.close_connection


def perform_sanity_checks(str_request) -> str:
    """ Performs a hash check on the message, and verifies the timestamp is valid.
        If either check fails, return False.  Otherwise, return True.
        Parameters:
            str_request: The request body as a string.
        Returns:
            True if the message is valid, False otherwise. """
    if settings.HASHKEY is not None:
        hash = security.get_message_hash(str_request)
        if settings.HASHKEY and hash not in str_request:
            print("Error: hash not match")
            return "Error: hash not match"
        if settings.HASHKEY and not security.verify_message_time(str_request):
            print("Error: timestamp expired")
            return "Error: timestamp expired"
    return ""


def main():
    """ This function starts the server. """
    webServer = HTTPServer(
        (settings.HOSTNAME, settings.SERVERPORT), IDGAFServer)
    print("Server started http://%s:%s" %
          (settings.HOSTNAME, settings.SERVERPORT))

    try:
        webServer.serve_forever()
    except KeyboardInterrupt:
        pass

    webServer.server_close()
    print("Server stopped.")


if __name__ == "__main__":
    """ This function starts the main method. """
    main()
